local BT_ACTION = BT_CUSTOM1
local BT_LIGHTDASH = BT_TOSSFLAG
local lskins = {
	["sonic"] = true,
	["tails"] = true,
	["knuckles"] = true,
	["amy"] = true,
	["fang"] = true,
	["metalsonic"] = true
}

local function ThokGhostThing(mobj)
	-- Thok object takes the sprite of it's target!
	if mobj.target then
		if (not (mobj.target.player and mobj.target.player.pflags & PF_STARTDASH)) 
			and (not (mobj.target.player and mobj.target.player.booming))
			and (not (mobj.target.player and mobj.target.player.jumpdiving)) then
			if mobj.target.player and (mobj.target.player.pflags & PF_SPINNING or mobj.target.player.lhoming) then -- after images only while spining
				mobj.skin = mobj.target.skin
				mobj.angle = mobj.target.angle
				mobj.sprite = mobj.target.sprite
				mobj.sprite2 = mobj.target.sprite2
				mobj.frame = mobj.target.frame
				mobj.frame = $ & ~FF_TRANSMASK
				mobj.frame = $|tr_trans50<<FF_TRANSSHIFT
			end
		end
	end
end

--
-- L_LookForEnemies
-- Looks for something you can hit - Used for homing attack
-- Includes monitors and springs!
--
local function L_LookForEnemies(player)
	local closestmo
	local an

	local range = 512*player.mo.scale
	searchBlockmap("objects", function(refmobj, mo)
		if not(mo.flags & MF_ENEMY or mo.flags & MF_BOSS or mo.flags & MF_MONITOR
			or mo.flags & MF_SPRING) then
			return -- not a valid enemy
		end

		if mo.health <= 0 then -- dead
			return
		end

		if mo == player.mo then
			return
		end

		if mo.flags2 & MF2_FRET then
			return
		end

		if mo.type == MT_DETON then
			return
		end
--[[
		if mo.flags & MF_MONITOR and mo.state == S_MONITOREXPLOSION5 then
			return
		end
]]
		if P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > range then
			return -- out of range
		end
			
			
		-- SRB2CB - homing fix	
		if mo.z < player.mo.z - range/2 then
			return -- too low
		end

		if mo.type == MT_PLAYER then -- Don't chase after other players!
			return
		end

		if closestmo and P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > P_AproxDistance(P_AproxDistance(player.mo.x-closestmo.x,
			player.mo.y-closestmo.y), player.mo.z-closestmo.z) then
			return
		end

		an = R_PointToAngle2(player.mo.x, player.mo.y, mo.x, mo.y) - player.mo.angle

		if AngleFixed(an) > 90*FRACUNIT and AngleFixed(an) < 270*FRACUNIT then
			return -- behind back
		end

		player.mo.angle = R_PointToAngle2(player.mo.x, player.mo.y, mo.x, mo.y)

		if not P_CheckSight(player.mo, mo)
			return
		end -- out of sight

		closestmo = mo
	end, player.mo, player.mo.x-range, player.mo.x+range, player.mo.y-range, player.mo.y+range)

	if closestmo then
		-- Found a target monster
		player.mo.target = closestmo
		player.mo.tracer = closestmo
		return true
	end

	return false
end


local function L_HomingAttack(source, enemy) -- Home in on your target
	local dist
	local dest

	if not enemy.health then
		return
	end

	if not source.tracer
		return -- Nothing to home in on!
	end

	-- adjust direction
	dest = source.tracer

	if not dest or dest.health <= 0 then
		return
	end

	-- change angle
	source.angle = R_PointToAngle2(source.x, source.y, enemy.x, enemy.y)
	if source.player then
		--[[if source.player == players[consoleplayer] then
			localangle = source.angle
		elseif cv_splitscreen.value and source.player == players[secondarydisplayplayer] then
			localangle2 = source.angle
		end]]
		source.player.cmd.angleturn = source.angle>>16
	end

	-- change slope
	dist = P_AproxDistance(P_AproxDistance(dest.x - source.x, dest.y - source.y),
		dest.z - source.z)

	if dist < 1 then
		dist = 1
	end

	if source.type == MT_DETON and enemy.player then -- For Deton Chase
		local ns = FixedDiv(enemy.player.normalspeed,(20*FRACUNIT)/17)
		source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), ns)
		source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), ns)
		source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), ns)
	elseif source.type ~= MT_PLAYER then
		if source.threshold == 32000 then
			local ns = source.info.speed/2
			source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), ns)
			source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), ns)
			source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), ns)
		else
			source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), source.info.speed)
			source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), source.info.speed)
			source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), source.info.speed)
		end
	elseif source.player then
		if (source.player.twodmove or twodlevel) and ((dest.y < source.player.mo.y+4*FRACUNIT) or (dest.y > source.player.mo.y-4*FRACUNIT)) then
			source.player.lhoming = 0
			return
		end
		
		source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), 60*FRACUNIT)
		if ((source.player and source.player.mo.flags2 & MF2_TWOD) or twodlevel) and enemy ~= MT_PLAYER then -- SRB2CBTODO: ?
			source.momy = 0
		else
			source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), 60*FRACUNIT)
		end
		source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), 60*FRACUNIT)
		if source.player.lhoming --[[and source.player.speed < 1]] then
			return
		end
	end
end

addHook("AbilitySpecial", function(player)
	if not (mapheaderinfo[gamemap].homing and lskins[player.mo.skin] and (player.charability == CA_THOK or player.charability == CA_HOMINGTHOK)) then
		return
	end

	if not (player.pflags & PF_THOKKED) or (player.charflags & SF_MULTIABILITY) then
		-- SRB2CB - 2D mode thok
		if (player.mo.flags2 & MF2_TWOD or twodlevel)--[[ and player.twodspeed >= 0]] then -- Less thok speed in 2D mode
			if player.mo.eflags & MFE_UNDERWATER then
				player.mo.momy = 0
				P_InstaThrustPlayer(player, player.mo.angle, (((player.normalspeed/FRACUNIT)<<FRACBITS)/2)--[[+player.twodspeed]])
			else
				player.mo.momy = 0
				P_InstaThrust(player.mo, player.mo.angle, ((player.normalspeed/FRACUNIT)<<FRACBITS)--[[+player.twodspeed]])
			end
		else
			-- Catapult the player
			if player.mo.eflags & MFE_UNDERWATER then
				P_InstaThrust(player.mo, player.mo.angle, ((player.actionspd/FRACUNIT)<<FRACBITS)/2)
			else
				P_InstaThrust(player.mo, player.mo.angle, (player.actionspd/FRACUNIT)<<FRACBITS)
			end
		end
		
		if player.mo.info.attacksound then S_StartSound(player.mo, player.mo.info.attacksound) end -- Play the THOK sound
		
		-- Now check the player's color so the right THOK object is displayed.
		if player.skincolor then
			local mobj
			local type
			
			if player.thokitem then
				type = player.thokitem
			else
				type = player.mo.info.painchance
			end
			
			mobj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z - (player.mo.info.height - player.mo.height)/3, type)
			mobj.color = player.mo.color
			mobj.target = player.mo
			ThokGhostThing(mobj)
		end
		
		if (--[[cv_homing.value or (player.charflags & SF_HOMING) or]] mapheaderinfo[gamemap].homing) and not player.lhoming and player.pflags & PF_JUMPED then
			if L_LookForEnemies(player) then
				if player.mo.tracer then
					player.lhoming = 3*TICRATE
				end
			end
		end
		--[[
		-- Hyper Sonic smart bomb
		if player.powers[pw_super] and (emeralds & EMERALD8) then
			player.blackow = 2
		end
		]]
		player.pflags = $ & ~PF_SPINNING
		player.pflags = $ & ~PF_STARTDASH
		player.pflags = $ | PF_THOKKED
	end

	return true
end)

addHook("PlayerThink", function(player)
	if not (mapheaderinfo[gamemap].homing and lskins[player.mo.skin]) then
		return
	end

	-- Uber-secret HOMING option. Experimental!
	if (mapheaderinfo[gamemap].homing and (player.charability == CA_THOK or player.charability == CA_HOMINGTHOK)) then
		player.charability = CA_THOK
		-- If you've got a target, chase after it!
		if player.lhoming and player.mo.tracer then
			local mobj
			local type
			
			if player.thokitem then
				type = player.thokitem
			else
				type = player.mo.info.painchance
			end
			
			mobj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z - (player.mo.info.height - player.mo.height)/3, type)
			--mobj.flags = (((player.powers[pw_super] and player.supercolor > 0) and player.supercolor or player.skincolor)<<MF_TRANSSHIFT)
			mobj.target = player.mo
			mobj.color = player.mo.color
			ThokGhostThing(mobj)
			
			L_HomingAttack(player.mo, player.mo.tracer)
			
			-- But if you don't, then stop homing.
			if player.mo.tracer.health <= 0 or (player.mo.tracer.flags2 & MF2_FRET)
				or (player.mo.state ~= S_PLAY_JUMP)
				or player.speed == 0
				or (player.mo.momz < 0 and P_IsObjectOnGround(player.mo))
				or (player.speed == 0 and (not (player.mo.momy and not player.mo.momx) or not (not player.mo.momy and player.mo.momx) or not (player.mo.momy and player.mo.momx and player.mo.momz) or not (not player.mo.momy and player.mo.momz) or not (player.mo.momy and not player.mo.momz) or not (not player.mo.momx and player.mo.momz) or not (player.mo.momx and not player.mo.momz))) then
				
				if player.lhoming and player.speed < FRACUNIT then -- stop homing, because you should be moving now
					return
				end
				
				-- When you're not moving while homing                        
                if player.speed == 0 and (not (player.mo.momy and not player.mo.momx) or not (not player.mo.momy and player.mo.momx) or not (player.mo.momy and player.mo.momx and player.mo.momz) or not (not player.mo.momy and player.mo.momz) or not (player.mo.momy and !player.mo.momz) or not (not player.mo.momx and player.mo.momz) or not (player.mo.momx and not player.mo.momz)) then                       
					return -- Stop then find the next homing point :
				end
				
				if player.mo.state ~= S_PLAY_JUMP
					or player.speed == 0
					or player.mo.momz < 0 and P_IsObjectOnGround(player.mo) then
					return
				end
				
				
				if player.mo.eflags & MFE_UNDERWATER then
					player.mo.momz = (457*FRACUNIT)/72
				else
					player.mo.momz = 10*FRACUNIT/NEWTICRATERATIO
				end
				
				player.mo.momx = 0
				player.mo.momy = 0
				player.lhoming = 0
				
				if player.mo.tracer.flags2 & MF2_FRET then
					P_InstaThrust(player.mo, player.mo.angle, -(player.speed/FRACUNIT <<(FRACBITS-3)))
				end
				
				if not (player.mo.tracer.flags & MF_BOSS) then
					player.pflags = $ & ~PF_THOKKED
				end
			end
		end
		
		-- If you're not jumping, then you obviously wouldn't be homing.
		if not (player.pflags & PF_JUMPED) then
			player.lhoming = 0
		end
	else
		player.lhoming = 0
	end

end)

addHook("MobjMoveCollide", function(mobj, spring)
	if not (mapheaderinfo[gamemap].homing and lskins[player.mo.skin] and (player.charability == CA_THOK or player.charability == CA_HOMINGTHOK)) then
		return
	end

	if not (mobj and mobj.valid and spring and spring.valid and mobj.player and mobj.player.valid) then
		return
	end
	if spring.flags & MF_SPRING then
		--[[if spring.info.damage or (maptol & TOL_ADVENTURE and mobj.player and mobj.player.homing) then -- Mimic SA
			if not spring.type ~= MT_YELLOWHORIZ or spring.type ~= MT_REDHORIZ or spring.type ~= MT_BUMPER then -- Don't restrict your speed, man
				mobj.momx = 0
				mobj.momy = 0
			end
			P_SetOrigin(mobj, spring.x, -- teleport the player to the middle of the spring,
	        	spring.y, mobj.z)-- so there's no issues with the angle he approaches at
		end]]
		if spring.type == MT_BALLOON and mobj.player.homing then
			mobj.player.mo.momy = $/2
			mobj.player.mo.momx = $/2
			mobj.player.homing = 0
		end

		--workaround (aqua)
		mobj.player.lhoming = 0
	end
end)

local function P_RingNearby(player) -- Is a ring in range?
	local closest
	local range = 192*player.mo.scale

	searchBlockmap("objects", function(refmobj, mo)
		if (mo.health <= 0) --[[or (mo.state == S_DISS)]] then -- Not a valid ring
			return
		end

		if mo.type ~= MT_RING then
			return
		end

		if P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > range then -- Out of range
			return
		end

		if not P_CheckSight(player.mo, mo) then -- Out of sight
			return
		end

		if closest and P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > P_AproxDistance(P_AproxDistance(player.mo.x-closest.x,
			player.mo.y-closest.y), player.mo.z-closest.z) then
			return
		end

		-- We don't wanna be dashing into hurt bosses repeatedly.
		if mo.flags2 & MF2_FRET then
			return
		end
		
		-- Found a target
		closest = mo
	end, player.mo, player.mo.x-range, player.mo.x+range, player.mo.y-range, player.mo.y+range)

	if closest then
		return true
	end

	return false;
end


local function P_LightDash(source, enemy) -- Home in on your target
	local dist
	local dest

	if not source.tracer then
		return -- Nothing to home in on!
	end

	-- adjust direction
	dest = source.tracer

	if not dest then
		return
	end

	-- change angle
	source.angle = R_PointToAngle2(source.x, source.y, enemy.x, enemy.y)
	if source.player then
		--[[
		if source.player == players[consoleplayer] then
			localangle = source.angle
		elseif cv_splitscreen.value and source.player == players[secondarydisplayplayer] then
			localangle2 = source.angle
		end
		]]
		source.player.cmd.angleturn = source.angle>>16
	end
	--R_PointToAngle2(source.z, source.z, enemy.z, enemy.z);

	-- change slope
	dist = P_AproxDistance(P_AproxDistance(dest.x - source.x, dest.y - source.y), dest.z - source.z)

	if (dist < 1)
		dist = 1
	end
		

 
    if source.player.speed < source.player.normalspeed then
		source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), (2*FRACUNIT+source.player.normalspeed))
		if source.player and not (source.player.mo.flags2 & MF2_TWOD) and not twodlevel then
			source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), (2*FRACUNIT+source.player.normalspeed))
		else
			source.momy = 0
		end
		--P_Thrust(source.player.mo, source.player.mo.angle, source.player.speed/FRACUNIT+1)
		source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), (40*FRACUNIT))
	else
		source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), (60*FRACUNIT/NEWTICRATERATIO))
		if source.player and not (source.player.mo.flags2 & MF2_TWOD) and not twodlevel then
			source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), (60*FRACUNIT/NEWTICRATERATIO))
		else
			source.momy = 0
		end
		--P_Thrust(source.player.mo, source.player.mo.angle, source.player.speed/FRACUNIT+1)
		source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), (40*FRACUNIT))
	end
end



local function P_LookForRings(player)
	local found = false
	
	player.mo.target = nil
	player.mo.tracer = nil
	local range = 256*player.mo.scale
	
	searchBlockmap("objects", function(refmobj, mo)
		if (mo.health <= 0) --[[or (mo.state == &states[S_DISS])]] then -- Not a valid ring
			return
		end
		
		if mo.type ~= MT_RING then
			return
		end
		
		if P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
						   player.mo.z-mo.z) > range then -- The higher this number is, the farther the game looks for the next waypoint
			return
		end
		
		if not P_CheckSight(player.mo, mo) then -- Out of sight
			return
		end
		
		if player.mo.target and P_AproxDistance(P_AproxDistance(player.mo.x-mo.x,
																player.mo.y-mo.y), player.mo.z-mo.z) >
			P_AproxDistance(P_AproxDistance(player.mo.x-player.mo.target.x,
											player.mo.y-player.mo.target.y), player.mo.z-player.mo.target.z) then
			return
		end
		
		-- We don't wanna be dashing into hurt bosses repeatedly.
		if mo.flags2 & MF2_FRET then
			return
		end
		
		-- Found a target
		found = true
		player.mo.target = mo
		player.mo.tracer = mo
	end, player.mo, player.mo.x-range, player.mo.x+range, player.mo.y-range, player.mo.y+range)
	
	if found then
		P_ResetPlayer(player)
		player.mo.state = S_PLAY_FALL
		P_ResetScore(player)
		P_LightDash(player.mo, player.mo.target)
		return
	end
	--player.mo.momx = FixedDiv(player.mo.momx,50*FRACUNIT) -- Change these for the momentum you keep after you finish flying
	--player.mo.momy = FixedDiv(player.mo.momy,50*FRACUNIT)
	--player.mo.momz = FixedDiv(player.mo.momz,50*FRACUNIT) -- 5 gives you just enough momentum
	player.llightdash = false
end

addHook("PlayerThink", function(player)

	if player.exiting or P_PlayerInPain(player)
	or (player.climbing or player.powers[pw_carry])
	or not lskins[player.mo.skin] then
		player.llightdashallowed = false
		player.llightdash = false
		player.lhoming = 0
		player.booming = false
		player.jumpdiving = false
		return
	end
	if P_IsObjectOnGround(player.mo) or not (player.pflags & PF_JUMPED) 
	or (gametyperules & GTR_TEAMFLAGS and player.gotflag)
	or player.pflags & PF_GLIDING or player.pflags & PF_SLIDING or player.pflags & PF_SHIELDABILITY then
		player.lhoming = 0
		player.jumpdiving = false
	end

	player.charflags = $|SF_RUNONWATER


	-- Light dashing
	if (--[[cv_lightdash.value or player.charflags & SF_LIGHTDASH or ]]mapheaderinfo[gamemap].lightdash) and P_RingNearby(player) and not player.powers[pw_flashing]
		--#ifndef ANGLE2D
		and not (player.powers[pw_carry] & CR_NIGHTSMODE)
--#endif
		then
		player.llightdashallowed = true
		
		if player.cmd.buttons & BT_LIGHTDASH then
			if not player.attackdown then
				player.llightdash = TICRATE
				player.attackdown = true
			end
		elseif not (player.cmd.buttons & BT_ATTACK) then
			player.attackdown = false
		end
	else
		player.llightdashallowed = false
	end
	
	if (--[[cv_lightdash.value or player.charflags & SF_LIGHTDASH or ]]mapheaderinfo[gamemap].lightdash) and player.llightdash 
		--#ifndef ANGLE2D
		and not (player.powers[pw_carry] & CR_NIGHTSMODE)
--#endif
		then
		if player.llightdash then
			P_LookForRings(player)
			-- player.powers[pw_flashing] = 2; -- This should only be included if flashing is turned to a value of int = 1
		end
	end

	-- Spin Disengage
	--	While you spin on the ground, you can get up and start running,
	-- for tight turns and such
	if (player.pflags & PF_SPINNING) and not (player.pflags & PF_STARTDASH) and P_IsObjectOnGround(player.mo) then
		if player.cmd.buttons & BT_ACTION then
			if leveltime % (TICRATE/5) == 0 then
				player.pflags = $ & ~PF_SPINNING
				player.mo.state = S_PLAY_WALK
				--topspeed = player.normalspeed
				P_InstaThrust(player.mo, player.mo.angle, player.speed)
			end
			
		end
	end
	
	if not P_IsObjectOnGround(player.mo) or player.pflags & PF_SPINNING or player.speed < 45*FRACUNIT or player.pflags & PF_STARTDASH
    or player.lhoming or player.llightdash or player.gliding or not (player.cmd.buttons & BT_ACTION) then
    	player.booming = 0
	end
  
	
	if player.mo.momz >= 0 or P_IsObjectOnGround(player.mo) or player.pflags & PF_SPINNING or player.pflags & PF_STARTDASH or player.lhoming or player.llightdash or player.gliding or not (player.cmd.buttons & BT_ACTION) then
		player.jumpdiving = false
	end
	
	if mapheaderinfo[gamemap].jumpdive and not player.powers[pw_super] then
		if (not P_IsObjectOnGround(player.mo) and (player.mo.z > player.mo.floorz + ((player.mo.flags2 & MF2_TWOD or twodlevel) and 32*FRACUNIT or 16*FRACUNIT)) and not player.llightdash and player.pflags & PF_JUMPED and not player.lhoming and not (player.pflags & PF_STARTDASH) and not player.gliding) and (player.mo.state == S_PLAY_JUMP or player.mo.state == S_PLAY_FALL) then
			if player.cmd.buttons & BT_SPIN and not (player.cmd.buttons & BT_JUMP) then
				player.mo.momz = $-3*NEWTICRATERATIO*FRACUNIT
				if not mapheaderinfo[gamemap].divemove then
					player.mo.momx = $/2
					player.mo.momy = $/2
				end
				-- Jumpdive graphics handled in P_MovePlayer
				if player.mo.momz < 0 then -- Make sure there's a trail only if you're going down
					player.jumpdiving = true
				end
		    end
		else
			player.jumpdiving = false
		end
	end
	
	
	
		
	if mapheaderinfo[gamemap].sonicboom then -- New mechanics for fast gameplay
		-- Sonic Boom	
		
		if P_IsObjectOnGround(player.mo) and not (player.pflags & PF_SPINNING) and not (player.pflags & PF_STARTDASH) then
			if player.cmd.buttons & BT_ACTION and ((player.mo.flags2 & MF2_TWOD or twodlevel) and not (player.cmd.sidemove > -1 and player.cmd.sidemove < 1) or player.cmd.forwardmove > 2) and (player.mo.momx ~= 0 or player.mo.momy ~= 0) then
				if (player.mo.flags2 & MF2_TWOD or twodlevel) then
					if player.speed < 3*(14*FRACUNIT+player.normalspeed)/2 then
						P_Thrust(player.mo, player.mo.angle, 5*FRACUNIT)
					end
				elseif player.speed < 60*FRACUNIT then
					--[[
					if player == players[consoleplayer] and cv_analog.value
					or (cv_splitscreen.value and player == players[secondarydisplayplayer] and cv_analog2.value) then
						if player.mo.flags2 & MF2_TWOSEVEN then
							--player.mo.angle = thiscam.angle
				       		P_Thrust(player.mo, player.mo.angle, 14*FRACUNIT+player.normalspeed)
				       	end
				    else
					   	P_InstaThrust(player.mo, player.mo.angle, 14*FRACUNIT+player.normalspeed)
					end
					]]
					if player.pflags & PF_ANALOGMODE then
						P_Thrust(player.mo, player.mo.angle, 14*FRACUNIT+player.normalspeed)
					else
						P_InstaThrust(player.mo, player.mo.angle, 14*FRACUNIT+player.normalspeed)
					end
				end
				
				if player.pflags & PF_SPINNING then
					player.mo.state = S_PLAY_WALK
					player.pflags = $ & ~PF_SPINNING
				end
				--topspeed = 14*FRACUNIT+player.normalspeed
				
				if (mapheaderinfo[gamemap].boomattract) and (mapheaderinfo[gamemap].noboomattack) then -- Attract rings and don't attack
					player.booming = 4
				elseif (mapheaderinfo[gamemap].boomattract) and (!mapheaderinfo[gamemap].noboomattack) then -- attract and attack
					player.booming = 3
				elseif mapheaderinfo[gamemap].noboomattack then -- don't attack
					player.booming = 2
				else
					player.booming = 1 -- attack, no attract
				end
				
-- Booming graphics handled elsewhere
				
			end
		end
	end







-- Show the "THOK!" graphic when spinning quickly across the ground,
	-- SRB2CB: or booming or jumpdiving
	if (player.skincolor and player.pflags & PF_SPINNING and player.speed > 15*FRACUNIT and not (player.pflags & PF_JUMPED)) or (player.booming and (player.speed > 20*FRACUNIT)) or player.jumpdiving then
		local mobj
		local zheight
		local type
		
		player.spinitem = 0
		if skins[player.mo.skin].spinitem then
			type = skins[player.mo.skin].spinitem ~= -1 and skins[player.mo.skin].spinitem or MT_THOK
		else
			type = player.mo.info.damage
		end
		
		zheight = player.mo.z	- (player.mo.info.height - player.mo.height)/3
		
		if zheight < player.mo.floorz then
			zheight = player.mo.floorz
		end
		
		mobj = P_SpawnMobj(player.mo.x, player.mo.y, zheight, type)
		--mobj.flags = ((player.powers[pw_super] and player.supercolor) and player.supercolor or player.skincolor)<<MF_TRANSSHIFT
		mobj.target = player.mo
		mobj.color = player.mo.color
		ThokGhostThing(mobj)
	end


	if player.lhoming then
		player.lhoming = $-1
	end
	if player.llightdash then
		player.llightdash = $-1
	end

	--#ifdef HWRENDER -- Eye candy
	
	if --[[rendermode ~= render_soft and rendermode ~= render_none 
		and ]]player.booming
		--[[and cv_grfovchange.value == 1]] then--player.speed > player.normalspeed and player.powers[pw_invulnerability] and !player.mfjumped and !netgame)
		-- When sonic booming, your normalspeed adjusts,
		-- you can't have the normalspeed this high otherwise, so it's ok to check this for booming
		local speed
		local startspeed
		
        startspeed = player.normalspeed
		
		speed = R_PointToDist2(player.mo.x + player.rmomx, player.mo.y + player.rmomy, player.mo.x, player.mo.y)
		
		if speed > (42)*FRACUNIT then
			speed = (42)*FRACUNIT
		end
		
		if speed >= startspeed then
			if not player.twodmove and not twodlevel then
				player.fovadd = speed+(speed/10) - startspeed
			end
		else
			player.fovadd = 0
		end
		
		if player.fovadd < 0 then
			player.fovadd = 0
		end
	else
		player.fovadd = 0
	end
	
	--#endif

end)

addHook("PlayerCanDamage", function(player, mobj)
	if (mapheaderinfo[gamemap].sonicboom and lskins[player.mo.skin])
	and player.booming == 1 or player.booming == 3 -- Do you possess the ability to subdue the object?
	and (mobj.flags & MF_BOSS) or ((mobj.flags & MF_ENEMY) and not (mobj.flags & MF_MISSILE)) then
		return true
	end
end)

hud.add(function(v, player)
	if not (mapheaderinfo[gamemap].lightdash and lskins[player.mo.skin]) then
		return
	end
	if player.llightdashallowed then
		v.drawFill(224, 6, 72, 12, 103|V_SNAPTOTOP|V_SNAPTORIGHT)
		v.drawString(260, 8, "ACTIVATE", V_TRANSLUCENT|V_SNAPTOTOP|V_SNAPTORIGHT, "center")
		v.drawString(260, 20, "LIGHT DASH", V_TRANSLUCENT|V_SNAPTOTOP|V_SNAPTORIGHT, "center")
	end
end, "game")
